#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_thesaurus.schema module

This module defines schemas interfaces and fields used o store thesaurus terms.
"""

from zope.interface import Interface, implementer
from zope.schema import Choice, List, Object, Set, TextLine
from zope.schema.interfaces import IList, IObject, SchemaNotProvided


__docformat__ = 'restructuredtext'


class IThesaurusTermField(IObject):
    """Marker interface to define a thesaurus term field"""

    thesaurus_name = TextLine(title="Thesaurus name",
                              required=False)

    extract_name = TextLine(title="Extract name",
                            required=False)


class IThesaurusTermsListField(IList):
    """Marker interface to define a list of thesaurus terms"""

    thesaurus_name = TextLine(title="Thesaurus name",
                              required=False)

    extract_name = TextLine(title="Extract name",
                            required=False)


@implementer(IThesaurusTermField)
class ThesaurusTermField(Object):
    """Thesaurus term schema field"""

    def __init__(self, schema=None, thesaurus_name='', extract_name='', **kw):
        super().__init__(schema=Interface, **kw)
        self.thesaurus_name = thesaurus_name
        self.extract_name = extract_name

    def _validate(self, value):
        super()._validate(value)
        # schema has to be provided by value
        if not self.schema.providedBy(value):
            raise SchemaNotProvided


@implementer(IThesaurusTermsListField)
class ThesaurusTermsListField(List):
    """Thesaurus terms list schema field"""

    def __init__(self, value_type=None, unique=False, thesaurus_name='', extract_name='', **kw):
        super().__init__(value_type=Object(schema=Interface), unique=False, **kw)
        self.thesaurus_name = thesaurus_name
        self.extract_name = extract_name


class ValidatedSet(Set):
    """A set field validated when not bound to a context"""

    def _validate(self, value):
        # Don't try to validate with empty context !
        if self.context is None:
            return
        super()._validate(value)


class ValidatedChoice(Choice):
    """An always validated choice field"""

    def _validate(self, value):
        """Don't need to validate!"""



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_thesaurus.index module

"""

__docformat__ = 'restructuredtext'

from ZODB.broken import Broken
from persistent import Persistent
from zope.interface import implementer
from zope.intid import IIntIds
from zope.schema.fieldproperty import FieldProperty

from pyams_catalog.index import KeywordIndexWithInterface
from pyams_thesaurus.interfaces.index import IThesaurusTermFieldIndex, \
    IThesaurusTermsListFieldIndex
from pyams_utils.registry import query_utility


_marker = object()


def get_terms(index, term):
    """Terms getter"""
    terms = [term, ]
    if index.include_parents:
        terms.extend(term.get_parents())
    if index.include_synonyms:
        if term.usage is not None:
            terms.append(term.usage)
        else:
            terms.extend(term.used_for)
    return terms


@implementer(IThesaurusTermFieldIndex)
class ThesaurusTermFieldIndex(KeywordIndexWithInterface):
    """Thesaurus term field index"""

    include_parents = FieldProperty(IThesaurusTermFieldIndex['include_parents'])
    include_synonyms = FieldProperty(IThesaurusTermFieldIndex['include_synonyms'])

    def __init__(self, interface, discriminator, family=None,
                 include_parents=False, include_synonyms=False):
        super().__init__(interface, discriminator, family)
        self.include_parents = include_parents
        self.include_synonyms = include_synonyms

    def discriminate(self, obj, default):
        if self.interface is not None:
            obj = self.interface(obj, None)
            if obj is None:
                return default

        if callable(self.discriminator):
            value = self.discriminator(obj, _marker)
        else:
            value = getattr(obj, self.discriminator, _marker)
            if callable(value):
                value = value(obj)

        if value is _marker:
            return default

        if value:
            intids = query_utility(IIntIds)
            value = set([
                intids.register(term)
                for term in get_terms(self, value)
            ])

        if isinstance(value, Persistent):
            raise ValueError('Catalog cannot index persistent object {0!r}'.format(value))

        if isinstance(value, Broken):
            raise ValueError('Catalog cannot index broken object {0!r}'.format(value))

        return value


@implementer(IThesaurusTermsListFieldIndex)
class ThesaurusTermsListFieldIndex(KeywordIndexWithInterface):
    """Thesaurus terms list field index"""

    include_parents = FieldProperty(IThesaurusTermsListFieldIndex['include_parents'])
    include_synonyms = FieldProperty(IThesaurusTermsListFieldIndex['include_synonyms'])

    def __init__(self, interface, discriminator, family=None,
                 include_parents=False, include_synonyms=False):
        super().__init__(interface, discriminator, family)
        self.include_parents = include_parents
        self.include_synonyms = include_synonyms

    def discriminate(self, obj, default):
        if self.interface is not None:
            obj = self.interface(obj, None)
            if obj is None:
                return default

        if callable(self.discriminator):
            value = self.discriminator(obj, _marker)
        else:
            value = getattr(obj, self.discriminator, _marker)
            if callable(value):
                value = value(obj)

        if value is _marker:
            return default

        if value:
            terms = []
            for term in value:
                terms.extend(get_terms(self, term))
            if terms:
                intids = query_utility(IIntIds)
                value = set([
                    intids.register(term)
                    for term in terms
                ])
        return value



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_thesaurus.manager module

This module defines a thesaurus "manager", which is just a container
utility to store thesaurus.
"""

__docformat__ = 'restructuredtext'

from zope.container.folder import Folder
from zope.schema.fieldproperty import FieldProperty

from pyams_thesaurus.interfaces.thesaurus import IThesaurusManager
from pyams_utils.factory import factory_config


@factory_config(IThesaurusManager)
class ThesaurusManager(Folder):
    """Thesaurus container"""

    show_home_menu = FieldProperty(IThesaurusManager['show_home_menu'])



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_thesaurus.interfaces.thesaurus module

This module defines thesaurus utility interfaces.
"""

from hypatia.interfaces import ICatalog
from zope.annotation import IAttributeAnnotatable
from zope.container.constraints import contains
from zope.container.interfaces import IContainer
from zope.interface import Attribute, Interface
from zope.location.interfaces import IContained
from zope.schema import Bool, Choice, Date, List, Object, Text, TextLine

from pyams_i18n.interfaces import BASE_LANGUAGES_VOCABULARY_NAME
from pyams_security.schema import PrincipalsSetField
from pyams_thesaurus.interfaces import THESAURUS_ADMIN_ROLE, THESAURUS_EXTRACTS_VOCABULARY, \
    THESAURUS_EXTRACT_MANAGER_ROLE, THESAURUS_MANAGER_ROLE, THESAURUS_NAMES_VOCABULARY
from pyams_thesaurus.interfaces.term import IThesaurusTerm
from pyams_utils.schema import ColorField


__docformat__ = 'restructuredtext'

from pyams_thesaurus import _  # pylint: disable=ungrouped-imports


class IThesaurusDescription(Interface):
    """Thesaurus base info"""

    title = TextLine(title=_("Title"),
                     description=_("Long title for this thesaurus"),
                     required=False)

    subject = TextLine(title=_("Subject"),
                       required=False)

    description = Text(title=_("Description"),
                       required=False)

    language = Choice(title=_("Language"),
                      description=_("Thesaurus's language"),
                      required=False,
                      default='en',
                      vocabulary=BASE_LANGUAGES_VOCABULARY_NAME)

    creator = TextLine(title=_("Creator"),
                       required=False)

    publisher = TextLine(title=_("Publisher"),
                         required=False)

    created = Date(title=_("Creation date"),
                   required=False)


class IThesaurusInfo(IThesaurusDescription):
    """Thesaurus base info"""

    name = TextLine(title=_("Thesaurus name"),
                    required=True)

    terms = Attribute("Thesaurus terms")

    top_terms = List(title=_("Thesaurus top-terms"),
                     description=_("List of top thesaurus terms, placed at first level"),
                     required=False,
                     value_type=Object(schema=IThesaurusTerm))

    catalog = Object(title=_("Thesaurus catalog"),
                     description=_("Inner thesaurus catalog, used for full-text indexing"),
                     schema=ICatalog)

    def init_catalog(self):
        """Initialize thesaurus catalog"""

    def load(self, configuration):
        """Load contents from given configuration"""

    def merge(self, configuration, thesaurus=None):
        """Merge current thesaurus with another one for given configuration"""

    def reset_terms_parent(self):
        """Reset thesaurus terms parent attribute"""

    def get_top_terms(self, extract=None):
        """Get top terms, for given extract or for the whole thesaurus"""

    def reset_top_terms(self):
        """Reset thesaurus top terms"""

    def clear(self):
        """Clear thesaurus contents"""

    def find_terms(self, query=None, extract=None, glob='end', limit=None,  # pylint: disable=too-many-arguments
                   exact=False, exact_only=False, stemmed=False):
        """Get terms matching given query and parent

        @param query: the text query
        @param extract: selected extract name, if any
        @param glob: can be 'start' (default), 'end', 'both' or None to handle globs
        @param limit: maximum number of search results
        @param exact: if True, terms exactly matching query will be returned
        @param exact_only: if False, stemmed query will also be executed
        @param stemmed: if True, query will use stemmed words
        """

    def get_nodes(self, term, result, subnodes=None, request=None):
        """Get tree nodes matching given term

        @param term: parent term
        @param result: output result from previous nodes, as method can be called recursively
        @param subnodes:
        """

    def delete(self):
        """Delete thesaurus"""


class IThesaurusRoles(Interface):
    """Thesaurus roles interface"""

    administrators = PrincipalsSetField(title=_("Administrators"),
                                        description=_("List of thesaurus's administrators"),
                                        role_id=THESAURUS_ADMIN_ROLE,
                                        required=False)

    managers = PrincipalsSetField(title=_("Contents managers"),
                                  description=_("List of thesaurus's contents contributors"),
                                  role_id=THESAURUS_MANAGER_ROLE,
                                  required=False)


class IThesaurus(IThesaurusInfo, IContained, IAttributeAnnotatable):
    """Thesaurus interface"""


THESAURUS_MANAGER_NAME = 'Thesaurus'


class IThesaurusManager(IContainer):
    """Thesaurus container utility interface"""

    contains(IThesaurus)

    show_home_menu = Bool(title=_("Access menu from home"),
                          description=_("If 'yes', a menu will be displayed to get access to "
                                        "thesaurus from site admin home page"),
                          required=True,
                          default=False)


class IThesaurusManagerTarget(Interface):
    """Marker interface for contents managing thesaurus"""


class IThesaurusTarget(Interface):
    """Marker interface for contents indexed on a thesaurus base"""


#
# Thesaurus extracts interfaces
#

class IThesaurusExtractInfo(Interface):
    """Thesaurus extract base info"""

    name = TextLine(title=_("Extract name"),
                    required=True)

    description = Text(title=_("Description"),
                       required=False)

    abbreviation = TextLine(title=_("Abbreviation"),
                            description=_("Short abbreviation used to distinguish the extract"),
                            required=True,
                            max_length=3)

    color = ColorField(title=_("Extract color"),
                       description=_("A color associated with this extract"),
                       required=True)

    def add_term(self, term):
        """Add a term to this extract"""

    def remove_term(self, term):
        """Remove a term from this extract"""

    def get_nodes(self, term, result, subnodes=None, request=None):
        """Extract terms getter"""


class IThesaurusExtractRoles(Interface):
    """Thesaurus extract roles"""

    managers = PrincipalsSetField(title=_("Extract managers"),
                                  description=_("List of principals which can manage extract "
                                                "contents"),
                                  role_id=THESAURUS_EXTRACT_MANAGER_ROLE,
                                  required=False)


class IThesaurusExtract(IThesaurusExtractInfo, IAttributeAnnotatable):
    """Thesaurus extract info"""


class IThesaurusExtracts(IContainer):
    """Thesaurus extracts container interface"""

    contains(IThesaurusExtractInfo)


class IThesaurusContextManager(Interface):
    """Thesaurus terms manager interface"""

    thesaurus_name = Choice(title=_("Thesaurus name"),
                            vocabulary=THESAURUS_NAMES_VOCABULARY,
                            required=False)

    extract_name = Choice(title=_("Thesaurus extract"),
                          vocabulary=THESAURUS_EXTRACTS_VOCABULARY,
                          required=False)


class IThesaurusContextManagerTarget(Interface):
    """Marker interface for tools managing thesaurus terms"""



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_thesaurus.interfaces.term module

This module defines thesaurus terms interfaces.
"""

from zope.container.interfaces import IContainer
from zope.interface import Attribute, Interface, Invalid, invariant
from zope.location.interfaces import IContained
from zope.schema import Bool, Choice, Datetime, Int, Set, Text, TextLine
from zope.schema.vocabulary import SimpleTerm, SimpleVocabulary

from pyams_thesaurus.interfaces.extension import THESAURUS_EXTENSIONS_VOCABULARY
from pyams_thesaurus.interfaces import THESAURUS_EXTRACTS_VOCABULARY
from pyams_thesaurus.schema import ThesaurusTermField, ThesaurusTermsListField, ValidatedSet


__docformat__ = 'restructuredtext'

from pyams_thesaurus import _


STATUS_CANDIDATE = 'candidate'
STATUS_PUBLISHED = 'published'
STATUS_ARCHIVED = 'archived'

THESAURUS_STATUS = (STATUS_CANDIDATE,
                    STATUS_PUBLISHED,
                    STATUS_ARCHIVED)

THESAURUS_STATUS_LABELS = (_("Candidate"),
                           _("Published"),
                           _("Archived"))

THESAURUS_STATUS_VOCABULARY = SimpleVocabulary([
    SimpleTerm(THESAURUS_STATUS[i], t, t)
    for i, t in enumerate(THESAURUS_STATUS_LABELS)
])


class IThesaurusTerm(IContained):
    """Thesaurus term base interface"""

    id = Attribute("Internal ID")

    label = TextLine(title=_("Term label"),
                     description=_("Full keyword for the given term"),
                     required=True)

    @invariant
    def check_label(self):
        """Check label for forbidden characters"""
        if '/' in self.label:  # pylint: disable=unsupported-membership-test
            raise Invalid(_("'/' character is forbidden in term's label"))

    base_label = Attribute("Base label without uppercase or accentuated character")

    title = Attribute("Term external title")

    alt = TextLine(title=_("Alternate label"),
                   description=_("Not to be confused with synonyms 'usage' label, given "
                                 "below..."),
                   required=False)

    definition = Text(title=_("Definition"),
                      description=_("Long definition, mostly for complicated terms"),
                      required=False)

    note = Text(title=_("Term's application note"),
                description=_("Application note for the given term"),
                required=False)

    generic = ThesaurusTermField(title=_("Generic term"),
                                 description=_("Parent generic term of the current term"),
                                 required=False)

    order = Int(title=_("Order"),
                description=_("Term's position between it's siblings; default terms order "
                              "is alphabetical"),
                required=False)

    specifics = ThesaurusTermsListField(title=_("Specifics terms"),
                                        description=_("Child more specifics terms of the "
                                                      "current term"),
                                        required=False)

    associations = ThesaurusTermsListField(title=_("Associated terms"),
                                           description=_("Other terms associated to the "
                                                         "current term"),
                                           required=False)

    usage = ThesaurusTermField(title=_("Usage"),
                               description=_("For synonyms, specify here the term's descriptor "
                                             "to use"),
                               required=False)

    used_for = ThesaurusTermsListField(title=_("Synonyms"),
                                       description=_("For a given allowed descriptor, specify "
                                                     "here the list of synonyms"),
                                       required=False)

    @invariant
    def check_synonym(self):
        """Check term for synonyms"""
        if self.usage is not None:
            if self.generic is not None:
                raise Invalid(_("A term can't be a synonym and attached to a generic term"))
            if self.used_for:
                raise Invalid(_("A term used as synonym can't have it's own synonyms "
                                "(all synonyms should be attached to the descriptor)"))

    extracts = ValidatedSet(title=_("Extracts"),
                            description=_("List of thesaurus extracts including this term"),
                            required=False,
                            value_type=Choice(vocabulary=THESAURUS_EXTRACTS_VOCABULARY))

    extensions = Set(title=_("Extensions"),
                     description=_("List of available extensions applied to this term"),
                     required=False,
                     value_type=Choice(vocabulary=THESAURUS_EXTENSIONS_VOCABULARY))

    status = Choice(title=_("Status"),
                    description=_("Term status"),
                    required=True,
                    vocabulary=THESAURUS_STATUS_VOCABULARY,
                    default='published')

    level = Attribute("Term's level in the thesaurus tree")

    micro_thesaurus = Bool(title=_("Micro-thesaurus?"),
                           description=_("Is the term part of a micro-thesaurus?"),
                           required=False)

    parent = ThesaurusTermField(title=_("First level parent"),
                                description=_("Parent at level 1 of the current term, or None"),
                                required=False,
                                schema=Interface)

    created = Datetime(title=_("Creation date"),
                       required=False)

    modified = Datetime(title=_("Modification date"),
                        required=False)

    def add_extract(self, extract, check=True):
        """Add given extract to the list of term extracts"""

    def remove_extract(self, extract, check=True):
        """Remove given extract from the list of term extracts"""

    def query_extensions(self):
        """Get list of extension utilities"""

    def get_parents(self):
        """Get list of term's parents"""

    def get_siblings(self):
        """Get siblings of current term"""

    def get_all_children(self, terms=None, with_synonyms=False):
        """Get full list of term's specifics"""

    def merge(self, term, configuration):
        """Merge term attributes with given term, to avoid overwriting all entity"""


IThesaurusTerm['generic'].schema = IThesaurusTerm
IThesaurusTerm['usage'].schema = IThesaurusTerm
IThesaurusTerm['parent'].schema = IThesaurusTerm


class IThesaurusLoaderTerm(Interface):
    """Marker interface for temporary thesaurus loader terms"""


class IThesaurusTermsContainer(IContainer):
    """Thesaurus terms container interface"""



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_thesaurus.loader.config module

"""

__docformat__ = 'restructuredtext'

from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty

from pyams_thesaurus.interfaces.loader import IThesaurusExporterConfiguration, \
    IThesaurusLoaderConfiguration, IThesaurusUpdaterConfiguration
from pyams_utils.factory import factory_config


@implementer(IThesaurusLoaderConfiguration)
class ThesaurusLoaderConfiguration:
    """Thesaurus loader configuration"""

    name = FieldProperty(IThesaurusLoaderConfiguration['name'])
    data = FieldProperty(IThesaurusLoaderConfiguration['data'])
    format = FieldProperty(IThesaurusLoaderConfiguration['format'])
    import_synonyms = FieldProperty(IThesaurusLoaderConfiguration['import_synonyms'])
    language = FieldProperty(IThesaurusLoaderConfiguration['language'])
    encoding = FieldProperty(IThesaurusLoaderConfiguration['encoding'])

    def __init__(self, data=None):
        if data:
            name = data.get('name')
            if name:
                self.name = name
            self.data = data.get('data')
            self.format = data.get('format')
            self.import_synonyms = data.get('import_synonyms')
            self.language = data.get('language')
            self.encoding = data.get('encoding')


@factory_config(IThesaurusUpdaterConfiguration)
class ThesaurusUpdaterConfiguration(ThesaurusLoaderConfiguration):
    """Thesaurus updater configuration"""

    clear = FieldProperty(IThesaurusUpdaterConfiguration['clear'])
    conflict_suffix = FieldProperty(IThesaurusUpdaterConfiguration['conflict_suffix'])

    def __init__(self, data=None):
        super().__init__(data)
        if data:
            self.clear = data.get('clear')
            self.conflict_suffix = data.get('conflict_suffix')


@factory_config(IThesaurusExporterConfiguration)
class ThesaurusExporterConfiguration:
    """Thesaurus exporter configuration"""

    filename = FieldProperty(IThesaurusExporterConfiguration['filename'])
    format = FieldProperty(IThesaurusExporterConfiguration['format'])
    extract = FieldProperty(IThesaurusExporterConfiguration['extract'])

    def __init__(self, data=None):
        if data:
            self.filename = data.get('filename')
            self.format = data.get('format')
            self.extract = data.get('extract')



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_thesaurus.zmi.tree module

This module defines components which are used to manage thesaurus terms tree.
"""

import json
from html import unescape

from pyramid.httpexceptions import HTTPBadRequest, HTTPForbidden, HTTPNotFound
from pyramid.response import Response
from pyramid.view import view_config
from zope.interface import Interface

from pyams_form.ajax import ajax_form_config
from pyams_form.button import Buttons, handler
from pyams_form.field import Fields
from pyams_form.interfaces.form import IAJAXFormRenderer
from pyams_layer.interfaces import IPyAMSLayer
from pyams_pagelet.pagelet import pagelet_config
from pyams_security.interfaces.base import VIEW_SYSTEM_PERMISSION
from pyams_skin.schema.button import CloseButton, SubmitButton
from pyams_skin.viewlet.menu import MenuItem
from pyams_template.template import template_config
from pyams_thesaurus.interfaces import MANAGE_THESAURUS_CONTENT_PERMISSION, \
    MANAGE_THESAURUS_EXTRACT_PERMISSION
from pyams_thesaurus.interfaces.extension import IThesaurusTermExtension
from pyams_thesaurus.interfaces.loader import IThesaurusExporter, IThesaurusExporterConfiguration, \
    IThesaurusLoader, IThesaurusUpdaterConfiguration
from pyams_thesaurus.interfaces.thesaurus import IThesaurus, IThesaurusExtracts
from pyams_thesaurus.loader.config import ThesaurusExporterConfiguration, \
    ThesaurusUpdaterConfiguration
from pyams_thesaurus.zmi.interfaces import IThesaurusTermsTreeMenu
from pyams_utils.adapter import ContextRequestViewAdapter, adapter_config
from pyams_utils.interfaces.tree import INode, ITree
from pyams_utils.registry import get_utility, query_utility
from pyams_utils.url import absolute_url
from pyams_viewlet.manager import viewletmanager_config
from pyams_viewlet.viewlet import ViewContentProvider, viewlet_config
from pyams_zmi.form import AdminModalAddForm
from pyams_zmi.interfaces import IAdminLayer, ICompositeView
from pyams_zmi.interfaces.viewlet import IActionsViewletManager, IContentManagementMenu
from pyams_zmi.view import CompositeAdminView
from pyams_zmi.zmi.viewlet.menu import NavigationMenuItem


__docformat__ = 'restructuredtext'

from pyams_thesaurus import _  # pylint: disable=ungrouped-imports


@viewletmanager_config(name='terms-tree.menu',
                       context=IThesaurus, layer=IAdminLayer,
                       manager=IContentManagementMenu, weight=20,
                       provides=IThesaurusTermsTreeMenu,
                       permission=VIEW_SYSTEM_PERMISSION)
class ThesaurusTermsTreeMenu(NavigationMenuItem):
    """Thesaurus terms tree menu"""

    label = _("Terms tree")
    icon_class = 'fas fa-tags'

    href = '#terms-view.html'


@pagelet_config(name='terms-view.html',
                context=IThesaurus, layer=IPyAMSLayer,
                permission=VIEW_SYSTEM_PERMISSION)
class ThesaurusTermsView(CompositeAdminView):
    """Thesaurus terms and extracts view"""


#
# Thesaurus terms tree view
#

@adapter_config(name='terms-tree.view',
                required=(IThesaurus, IAdminLayer, ThesaurusTermsView),
                provides=ICompositeView)
@template_config(template='templates/terms-tree.pt', layer=IAdminLayer)
class ThesaurusTermsTreeView(ViewContentProvider):
    """Thesaurus terms tree view"""

    title = _("Terms tree view")

    weight = 20

    @property
    def extracts(self):
        """Thesaurus extracts getter"""
        return IThesaurusExtracts(self.context)

    @property
    def tree(self):
        """Terms tree root nodes getter"""
        return [
            INode(node)
            for node in ITree(self.context).get_root_nodes()
        ]

    @property
    def search_query_params(self):
        """Terms search query params getter"""
        return json.dumps({
            'thesaurus_name': self.context.name
        })


@view_config(name='get-nodes.json',
             context=IThesaurus, request_type=IPyAMSLayer,
             permission=VIEW_SYSTEM_PERMISSION,
             renderer='json', xhr=True)
def get_thesaurus_nodes(request):
    """Get thesaurus nodes"""
    label = request.params.get('term')
    if label:
        label = unescape(label)
    term = request.context.terms.get(label)
    if term is None:
        raise HTTPNotFound
    result = []
    request.context.get_nodes(term, result)
    return {
        'term': label,
        'nodes': result
    }


@view_config(name='get-parent-nodes.json',
             context=IThesaurus, request_type=IPyAMSLayer,
             permission=VIEW_SYSTEM_PERMISSION,
             renderer='json', xhr=True)
def get_parent_nodes(request):
    """Get thesaurus parent nodes of a given node"""
    label = request.params.get('term')
    if label:
        label = unescape(label)
    term = request.context.terms.get(label)
    if term is None:
        raise HTTPNotFound
    result = []
    parents = list(reversed(term.get_parents()))
    if parents:
        request.context.get_nodes(parents[0], result, [t.label for t in parents])
        return {
            'term': label,
            'nodes': result,
            'parent': parents[0].label
        }
    return {
        'term': label,
        'nodes': result,
        'parent': label
    }


@view_config(name='switch-extension.json',
             context=IThesaurus, request_type=IPyAMSLayer,
             permission=MANAGE_THESAURUS_CONTENT_PERMISSION,
             renderer='json', xhr=True)
def switch_term_extension(request):
    """Term extension switcher"""
    label = request.params.get('term')
    extension_name = request.params.get('extension')
    if not (label and extension_name):
        raise HTTPBadRequest("Missing argument(s)")
    thesaurus = request.context
    term = thesaurus.terms.get(unescape(label))
    if term is None:
        raise HTTPBadRequest("Term not found")
    extension = query_utility(IThesaurusTermExtension, name=extension_name)
    if extension is None:
        raise HTTPBadRequest("Extension not found")
    term.extensions = term.extensions | {extension_name}
    translate = request.localizer.translate
    return {
        'name': extension_name,
        'title': translate(extension.label),
        'icon': extension.icon_css_class,
        'view': absolute_url(term, request, extension.target_view),
        'active': extension_name in term.extensions
    }


@view_config(name='switch-extract.json',
             context=IThesaurus, request_type=IPyAMSLayer,
             permission=VIEW_SYSTEM_PERMISSION,
             renderer='json', xhr=True)
def switch_term_extract(request):
    """Term extract switcher"""
    label = request.params.get('term')
    extract_name = request.params.get('extract')
    if not (label and extract_name):
        raise HTTPBadRequest("Missing argument(s)")
    thesaurus = request.context
    term = thesaurus.terms.get(unescape(label))
    if term is None:
        raise HTTPBadRequest("Term not found")
    extract = IThesaurusExtracts(thesaurus).get(extract_name)
    if extract is None:
        raise HTTPBadRequest("Extract not found")
    if not request.has_permission(MANAGE_THESAURUS_EXTRACT_PERMISSION, context=extract):
        raise HTTPForbidden("You are not authorized to update this extract")
    if extract.name in (term.extracts or ()):
        extract.remove_term(term)
    else:
        extract.add_term(term)
    return {
        'term': term.label,
        'extract': extract.name,
        'color': extract.color,
        'used': extract.name in term.extracts
    }


#
# Import menu
#

@viewlet_config(name='import-terms.action',
                context=IThesaurus, layer=IAdminLayer, view=ThesaurusTermsTreeView,
                manager=IActionsViewletManager, weight=10,
                permission=MANAGE_THESAURUS_CONTENT_PERMISSION)
class ThesaurusImportAction(MenuItem):
    """Thesaurus import action"""

    label = _("Import terms")
    icon_class = 'fas fa-upload'

    href = 'import-terms.html'
    modal_target = True


class IThesaurusImportFormButtons(Interface):
    """Thesaurus import form buttons interface"""

    add = SubmitButton(name='add',
                       title=_("Import terms"))

    close = CloseButton(name='close',
                        title=_("Cancel"))


@ajax_form_config(name='import-terms.html',
                  context=IThesaurus, layer=IPyAMSLayer,
                  permission=MANAGE_THESAURUS_CONTENT_PERMISSION)
class ThesaurusImportForm(AdminModalAddForm):
    """Thesaurus import form"""

    @property
    def title(self):
        """Title getter"""
        translate = self.request.localizer.translate
        return '<small>{}</small><br />{}'.format(
            translate(_("Thesaurus: {}")).format(self.context.name),
            translate(_("Import new terms")))

    legend = _("Import thesaurus terms")

    fields = Fields(IThesaurusUpdaterConfiguration).select('clear', 'conflict_suffix', 'data',
                                                           'format', 'import_synonyms',
                                                           'language', 'encoding')
    buttons = Buttons(IThesaurusImportFormButtons)

    def update_widgets(self, prefix=None):
        super().update_widgets(prefix)
        language = self.widgets.get('language')
        if language is not None:
            language.no_value_message = _("-- automatic detection -- (if available)")
        encoding = self.widgets.get('encoding')
        if encoding is not None:
            encoding.no_value_message = _("-- automatic detection -- (if available)")

    @handler(IThesaurusImportFormButtons['add'])
    def handle_import(self, action):
        """Import action handler"""
        super().handle_add(self, action)  # pylint: disable=too-many-function-args

    def create(self, data):
        return ThesaurusUpdaterConfiguration()

    def add(self, config):  # pylint: disable=arguments-renamed
        loader = get_utility(IThesaurusLoader, name=config.format)
        source = loader.load(config.data, config)
        target = self.context
        if config.clear:
            target.clear()
        target.merge(config, source)


@adapter_config(required=(IThesaurus, IAdminLayer, ThesaurusImportForm),
                provides=IAJAXFormRenderer)
class ThesaurusImportFormRenderer(ContextRequestViewAdapter):
    """Thesaurus import form renderer"""

    def render(self, changes):  # pylint: disable=no-self-use
        """AJAX form renderer"""
        if changes is None:
            return None
        return {
            'status': 'reload'
        }


#
# Thesaurus export form
#

@viewlet_config(name='export-terms.action',
                context=IThesaurus, layer=IAdminLayer, view=ThesaurusTermsTreeView,
                manager=IActionsViewletManager, weight=20,
                permission=MANAGE_THESAURUS_CONTENT_PERMISSION)
class ThesaurusExportAction(MenuItem):
    """Thesaurus export action"""

    label = _("Export terms")
    icon_class = 'fas fa-download'

    href = 'export-terms.html'
    modal_target = True


class IThesaurusExportFormButtons(Interface):
    """Thesaurus export form buttons interface"""

    export = SubmitButton(name='export',
                          title=_("terms-export-button", default="Export terms"))

    close = CloseButton(name='close',
                        title=_("Cancel"))


@ajax_form_config(name='export-terms.html',  # pylint: disable=abstract-method
                  context=IThesaurus, layer=IPyAMSLayer,
                  permission=MANAGE_THESAURUS_CONTENT_PERMISSION,
                  ajax_xhr=False, ajax_require_csrf=False)
class ThesaurusExportForm(AdminModalAddForm):
    """Thesaurus export form"""

    @property
    def title(self):
        """Title getter"""
        translate = self.request.localizer.translate
        return '<small>{}</small><br />{}'.format(
            translate(_("Thesaurus: {}")).format(self.context.name),
            translate(_("terms-export-title", default="Export terms")))

    legend = _("Export thesaurus terms")

    fields = Fields(IThesaurusExporterConfiguration)
    buttons = Buttons(IThesaurusExportFormButtons)

    object_data = {
        'ams-form-download-target': 'thesaurus-export-frame'
    }

    exporter = None

    @handler(IThesaurusExportFormButtons['export'])
    def handle_export(self, action):
        """Export action handler"""
        super().handle_add(self, action)  # pylint: disable=too-many-function-args

    def create_and_add(self, data):
        data = data.get(self, {})
        config = ThesaurusExporterConfiguration(data)
        self.exporter = get_utility(IThesaurusExporter, name=config.format)
        target = self.exporter.export(self.context, config)
        return target, config


@adapter_config(required=(IThesaurus, IAdminLayer, ThesaurusExportForm),
                provides=IAJAXFormRenderer)
class ThesaurusExportFormRenderer(ContextRequestViewAdapter):
    """Thesaurus export form renderer"""

    def render(self, changes):
        """AJAX form renderer"""
        if changes is None:
            return None
        target, config = changes
        target.seek(0)
        headers = {
            'Content-Disposition': 'attachment; filename="{0}"'.format(config.filename),
            'Content-Type': self.view.exporter.handler.content_type
        }
        response = Response(headers=headers)
        response.body_file = target
        return response



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS thesaurus.include module

This module is used for Pyramid integration.
"""

import re

from pyams_security.interfaces.base import ROLE_ID
from pyams_security.interfaces.names import ADMIN_USER_ID, SYSTEM_ADMIN_ROLE
from pyams_thesaurus.interfaces import ADMIN_THESAURUS_PERMISSION, CREATE_THESAURUS_PERMISSION, \
    MANAGE_THESAURUS_CONTENT_PERMISSION, MANAGE_THESAURUS_EXTRACT_PERMISSION, \
    REST_EXTRACTS_GETTER_ROUTE, REST_TERMS_SEARCH_ROUTE, THESAURUS_ADMIN_ROLE, \
    THESAURUS_EXTRACT_MANAGER_ROLE, \
    THESAURUS_MANAGER_ROLE


__docformat__ = 'restructuredtext'

from pyams_thesaurus import _  # pylint: disable=ungrouped-imports


def include_package(config):
    """Pyramid package include"""

    # add translations
    config.add_translation_dirs('pyams_thesaurus:locales')

    # register permissions
    config.register_permission({
        'id': CREATE_THESAURUS_PERMISSION,
        'title': _("Create thesaurus")
    })
    config.register_permission({
        'id': ADMIN_THESAURUS_PERMISSION,
        'title': _("Manage thesaurus properties")
    })
    config.register_permission({
        'id': MANAGE_THESAURUS_CONTENT_PERMISSION,
        'title': _("Manage thesaurus content")
    })
    config.register_permission({
        'id': MANAGE_THESAURUS_EXTRACT_PERMISSION,
        'title': _("Manage thesaurus extract")
    })

    # upgrade system manager role
    config.upgrade_role(SYSTEM_ADMIN_ROLE,
                        permissions={
                            ADMIN_THESAURUS_PERMISSION,
                            MANAGE_THESAURUS_CONTENT_PERMISSION,
                            MANAGE_THESAURUS_EXTRACT_PERMISSION
                        })

    # register roles
    config.register_role({
        'id': THESAURUS_ADMIN_ROLE,
        'title': _("Thesaurus administrator (role)"),
        'permissions': {
            ADMIN_THESAURUS_PERMISSION,
            MANAGE_THESAURUS_CONTENT_PERMISSION,
            MANAGE_THESAURUS_EXTRACT_PERMISSION
        },
        'managers': {
            ADMIN_USER_ID,
            ROLE_ID.format(SYSTEM_ADMIN_ROLE)
        }
    })
    config.register_role({
        'id': THESAURUS_MANAGER_ROLE,
        'title': _("Thesaurus content manager (role)"),
        'permissions': {
            MANAGE_THESAURUS_CONTENT_PERMISSION,
            MANAGE_THESAURUS_EXTRACT_PERMISSION
        },
        'managers': {
            ADMIN_USER_ID,
            ROLE_ID.format(SYSTEM_ADMIN_ROLE),
            ROLE_ID.format(THESAURUS_ADMIN_ROLE)
        }
    })
    config.register_role({
        'id': THESAURUS_EXTRACT_MANAGER_ROLE,
        'title': _("Thesaurus extract manager (role)"),
        'permissions': {
            MANAGE_THESAURUS_EXTRACT_PERMISSION
        },
        'managers': {
            ADMIN_USER_ID,
            ROLE_ID.format(SYSTEM_ADMIN_ROLE),
            ROLE_ID.format(THESAURUS_ADMIN_ROLE)
        }
    })

    # register REST API routes
    config.add_route(REST_TERMS_SEARCH_ROUTE,
                     config.registry.settings.get(f'{REST_TERMS_SEARCH_ROUTE}_route',
                                                  '/api/thesaurus/terms'))
    config.add_route(REST_EXTRACTS_GETTER_ROUTE,
                     config.registry.settings.get(f'{REST_EXTRACTS_GETTER_ROUTE}_route',
                                                  '/api/thesaurus/extracts'))

    # test Venusian scans
    ignored = []

    try:
        import pyams_zmi  # pylint: disable=import-outside-toplevel,unused-import
    except ImportError:
        ignored.append(re.compile(r'pyams_thesaurus\..*\.zmi\.?.*').search)
    else:
        if not config.registry.settings.get('pyams_thesaurus.include_tests_features', False):
            ignored.append('pyams_thesaurus.zmi.test')

    config.scan(ignore=ignored)



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_thesaurus.term module

This module defines main thesaurus terms class.
"""

from datetime import datetime

from persistent import Persistent
from pyramid.events import subscriber
from zope.container.contained import Contained
from zope.interface import alsoProvides, noLongerProvides
from zope.interface.interfaces import ComponentLookupError
from zope.lifecycleevent import IObjectAddedEvent, IObjectModifiedEvent, IObjectMovedEvent, \
    IObjectRemovedEvent
from zope.schema.fieldproperty import FieldProperty
from zope.schema.vocabulary import SimpleTerm, SimpleVocabulary
from zope.traversing.interfaces import ITraversable

from pyams_catalog.utils import index_object, reindex_object, unindex_object
from pyams_security.interfaces import IViewContextPermissionChecker
from pyams_thesaurus.interfaces import MANAGE_THESAURUS_CONTENT_PERMISSION
from pyams_thesaurus.interfaces.extension import IThesaurusTermExtension, \
    THESAURUS_EXTENSIONS_VOCABULARY
from pyams_thesaurus.interfaces.term import IThesaurusLoaderTerm, IThesaurusTerm, \
    IThesaurusTermsContainer, STATUS_PUBLISHED
from pyams_thesaurus.interfaces.thesaurus import IThesaurus, IThesaurusExtract
from pyams_utils.adapter import ContextAdapter, adapter_config
from pyams_utils.factory import factory_config
from pyams_utils.interfaces.tree import INode
from pyams_utils.registry import get_pyramid_registry, query_utility
from pyams_utils.request import check_request
from pyams_utils.timezone import tztime
from pyams_utils.traversing import get_parent
from pyams_utils.unicode import translate_string
from pyams_utils.vocabulary import vocabulary_config


__docformat__ = 'restructuredtext'


REVERSE_LINK_ATTRIBUTES = {
    'generic': 'specifics',
    'usage': 'used_for'
}

REVERSE_LIST_ATTRIBUTES = {
    'specifics': 'generic',
    'used_for': 'usage'
}


@adapter_config(name='terms',
                required=IThesaurus,
                provides=ITraversable)
class ThesaurusTermsNamespace(ContextAdapter):
    """Thesaurus ++terms++ namespace"""

    def traverse(self, name, furtherpath=None):  # pylint: disable=unused-argument
        """Namespace traverser"""
        terms = IThesaurus(self.context).terms
        if name:
            return terms[name]
        return terms


@factory_config(IThesaurusTerm)  # pylint: disable=too-many-instance-attributes
class ThesaurusTerm(Persistent, Contained):
    """Thesaurus term"""

    label = FieldProperty(IThesaurusTerm['label'])
    alt = FieldProperty(IThesaurusTerm['alt'])
    definition = FieldProperty(IThesaurusTerm['definition'])
    note = FieldProperty(IThesaurusTerm['note'])
    _generic = FieldProperty(IThesaurusTerm['generic'])
    order = FieldProperty(IThesaurusTerm['order'])
    _specifics = FieldProperty(IThesaurusTerm['specifics'])
    _associations = FieldProperty(IThesaurusTerm['associations'])
    _usage = FieldProperty(IThesaurusTerm['usage'])
    _used_for = FieldProperty(IThesaurusTerm['used_for'])
    _extracts = FieldProperty(IThesaurusTerm['extracts'])
    _extensions = FieldProperty(IThesaurusTerm['extensions'])
    status = FieldProperty(IThesaurusTerm['status'])
    micro_thesaurus = FieldProperty(IThesaurusTerm['micro_thesaurus'])
    parent = FieldProperty(IThesaurusTerm['parent'])
    _created = FieldProperty(IThesaurusTerm['created'])
    _modified = FieldProperty(IThesaurusTerm['modified'])

    def __init__(self, label, alt=None, definition=None, note=None,  # pylint: disable=too-many-arguments,too-many-instance-attributes
                 generic=None, specifics=None, associations=None,
                 usage=None, used_for=None, created=None, modified=None):
        self.label = label
        self.alt = alt
        self.definition = definition
        self.note = note
        self.generic = generic
        self.specifics = specifics or []
        self.associations = associations or []
        self.usage = usage
        self.used_for = used_for or []
        self.created = created
        self.modified = modified

    def __eq__(self, other):
        if other is None:
            return False
        return isinstance(other, ThesaurusTerm) and (self.label == other.label)

    def __hash__(self):
        return hash(self.label)

    @property
    def base_label(self):
        """Get translated label"""
        return translate_string(self.label, escape_slashes=True, force_lower=True, spaces=' ')

    @property
    def title(self):
        """Complete title getter"""
        if self._usage:
            label = self._usage.label
            terms = [
                term.label for term in self._usage.used_for
                if term.status == STATUS_PUBLISHED
            ]
        elif self._used_for:
            label = self.label
            terms = [
                term.label for term in self._used_for
                if term.status == STATUS_PUBLISHED
            ]
        else:
            label = self.label
            terms = []
        return label + (' [ {0} ]'.format(', '.join(terms)) if terms else '')

    @property
    def generic(self):
        """Generic term getter"""
        return self._generic

    @generic.setter
    def generic(self, value):
        """Generic term setter"""
        self._generic = value
        if value is not None:
            self.extracts = self.extracts & value.extracts

    @property
    def specifics(self):
        """Specific terms getter"""
        return sorted(self._specifics,
                      key=lambda x: (x.order or 999, x.label))

    @specifics.setter
    def specifics(self, value):
        """Specific terms setter"""
        self._specifics = list(value or ())

    @property
    def associations(self):
        """Associated terms getter"""
        return self._associations

    @associations.setter
    def associations(self, value):
        """Associated terms setter"""
        self._associations = list(value or ())

    @property
    def usage(self):
        """Usage term getter"""
        return self._usage

    @usage.setter
    def usage(self, value):
        """Usage term setter"""
        self._usage = value
        if value is not None:
            self.generic = None
            self.extracts = value.extracts

    @property
    def used_for(self):
        """Used for term getter"""
        return self._used_for

    @used_for.setter
    def used_for(self, value):
        """Used for term setter"""
        self._used_for = list(value or ())

    @property
    def extracts(self):
        """Term extracts getter"""
        return self._extracts or set()

    @extracts.setter
    def extracts(self, value):
        """Term extracts setter"""
        old_value = self._extracts or set()
        new_value = value or set()
        if self._generic is not None:
            new_value = new_value & (self._generic.extracts or set())
        if old_value != new_value:
            removed = old_value - new_value
            if removed:
                for term in self.specifics:
                    term.extracts = (term.extracts or set()) - removed
            self._extracts = new_value
            # Extracts selection also applies to term synonyms...
            for term in self.used_for or ():
                term.extracts = self.extracts

    def add_extract(self, extract, check=True):
        """Add term to given extract"""
        if IThesaurusExtract.providedBy(extract):  # pylint: disable=no-value-for-parameter
            extract = extract.name
        if check:
            self.extracts = (self.extracts or set()) | {extract}
        else:
            self._extracts = (self._extracts or set()) | {extract}
            # Extracts selection also applies to term synonyms...
            for term in self.used_for or ():
                term.extracts = self.extracts

    def remove_extract(self, extract, check=True):
        """Removed term from given extract"""
        if IThesaurusExtract.providedBy(extract):  # pylint: disable=no-value-for-parameter
            extract = extract.name
        if check:
            self.extracts = (self.extracts or set()) - {extract}
        else:
            self._extracts = (self._extracts or set()) - {extract}
            # Extracts selection also applies to term synonyms...
            for term in self.used_for or ():
                term.extracts = self.extracts

    @property
    def extensions(self):
        """Term extensions getter"""
        return self._extensions or set()

    @extensions.setter
    def extensions(self, value):
        """Term extensions setter"""
        old_value = self._extensions or set()
        new_value = value or set()
        if old_value != new_value:
            added = new_value - old_value
            removed = old_value - new_value
            for ext in removed:
                extension = query_utility(IThesaurusTermExtension, ext)
                if extension is not None:
                    noLongerProvides(self, extension.target_interface)
            for ext in added:
                extension = query_utility(IThesaurusTermExtension, ext)
                if extension is not None:
                    alsoProvides(self, extension.target_interface)
            self._extensions = new_value

    def query_extensions(self):
        """Query term extensions"""
        return sorted([
            util for util in [
                query_utility(IThesaurusTermExtension, ext)
                for ext in self.extensions
            ]
            if util is not None
        ], key=lambda x: getattr(x, 'weight', 999))

    @property
    def created(self):
        """Creation date getter"""
        return self._created

    @created.setter
    def created(self, value):
        """Creation date setter"""
        if isinstance(value, str):
            if ' ' in value:
                value = datetime.strptime(value, '%Y-%m-%d %H:%M:%S')
            else:
                value = datetime.strptime(value, '%Y-%m-%d')
        self._created = tztime(value)

    @property
    def modified(self):
        """Modified date getter"""
        return self._modified

    @modified.setter
    def modified(self, value):
        """Modified date setter"""
        if isinstance(value, str):
            if ' ' in value:
                value = datetime.strptime(value, '%Y-%m-%d %H:%M:%S')
            else:
                value = datetime.strptime(value, '%Y-%m-%d')
        self._modified = tztime(value)

    def get_parents(self):
        """Term parents getter"""
        terms = []
        parent = self.generic
        while parent is not None:
            terms.append(parent)
            parent = parent.generic
        return terms

    @property
    def level(self):
        """Term depth level getter"""
        return len(self.get_parents()) + 1

    def get_siblings(self):
        """Term siblings getter"""
        terms = []
        parent = self.generic
        if parent is not None:
            for term in parent.specifics:
                if term is not self:
                    terms.append(term)
        return terms

    def get_all_children(self, terms=None, with_synonyms=False):
        """Term children getter"""
        if terms is None:
            terms = []
        if with_synonyms:
            terms.extend(self.used_for)
        terms.extend(self.specifics)
        for term in self.specifics:
            term.get_all_children(terms, with_synonyms)
        return terms

    def short_url(self):
        """Short URL getter"""
        thesaurus = get_parent(self, IThesaurus)
        return '{}/{}'.format(thesaurus.__name__, self.__name__)

    def merge(self, term, configuration):  # pylint: disable=too-many-branches,too-many-statements
        """Merge existing term with provided one"""
        # terms marked by IThesaurusLoaderTerm interface are used by custom loaders which
        # only contains synonyms definitions; so they shouldn't alter terms properties
        if term is None:
            return
        # assign basic attributes
        if not IThesaurusLoaderTerm.providedBy(term):  # pylint: disable=no-value-for-parameter
            for name in ('label', 'definition', 'note', 'status', 'micro_thesaurus',
                         'created', 'modified'):
                setattr(self, name, getattr(term, name, None))
        # for term references, we have to check if the target term is already
        # in our parent thesaurus or not :
        #  - if yes => we target the term actually in the thesaurus
        #  - if not => we keep the same target, which will be included in the thesaurus
        #  after merging
        terms = self.__parent__
        if IThesaurusLoaderTerm.providedBy(term):  # pylint: disable=no-value-for-parameter
            attrs = ('usage',)
        else:
            attrs = ('generic', 'usage')
        for name in attrs:
            target = getattr(term, name)
            if target is None:
                setattr(self, name, None)
            else:
                label = target.label
                suffix = getattr(configuration, 'conflict_suffix', None)
                if suffix:
                    label = target.label + ' ' + suffix
                    if label not in terms:  # pylint: disable=unsupported-membership-test
                        label = target.label
                if label in terms:  # pylint: disable=unsupported-membership-test
                    target_term = terms[label]  # pylint: disable=unsubscriptable-object
                else:
                    target_term = target
                setattr(self, name, target_term)
                if name in REVERSE_LINK_ATTRIBUTES:
                    attribute = REVERSE_LINK_ATTRIBUTES[name]
                    setattr(target_term, attribute, set(getattr(target_term, attribute)) | {self})
        if IThesaurusLoaderTerm.providedBy(term):  # pylint: disable=no-value-for-parameter
            attrs = ('used_for',)
        else:
            attrs = ('specifics', 'associations', 'used_for')
        for name in attrs:
            targets = getattr(term, name, [])
            if not targets:
                setattr(self, name, [])
            else:
                new_targets = []
                for target in targets:
                    label = target.label
                    suffix = getattr(configuration, 'conflict_suffix', None)
                    if suffix:
                        label = target.label + ' ' + suffix
                        if label not in terms:  # pylint: disable=unsupported-membership-test
                            label = target.label
                    if label in terms:  # pylint: disable=unsupported-membership-test
                        target_term = terms[label]  # pylint: disable=unsubscriptable-object
                    else:
                        target_term = target
                    new_targets.append(target_term)
                    if name in REVERSE_LIST_ATTRIBUTES:
                        attribute = REVERSE_LIST_ATTRIBUTES[name]
                        setattr(target_term, attribute, self)
                setattr(self, name, new_targets)


@subscriber(IObjectAddedEvent, context_selector=IThesaurusTerm)
@subscriber(IObjectMovedEvent, context_selector=IThesaurusTerm)
def handle_new_term(event):
    """Index term into inner catalog"""
    if IThesaurusLoaderTerm.providedBy(event.object):  # pylint: disable=no-value-for-parameter
        return
    if IThesaurusTermsContainer.providedBy(event.oldParent):  # pylint: disable=no-value-for-parameter
        thesaurus = event.oldParent.__parent__
        if IThesaurus.providedBy(thesaurus):  # pylint: disable=no-value-for-parameter
            unindex_object(event.object, thesaurus.catalog)
    if IThesaurusTermsContainer.providedBy(event.newParent):  # pylint: disable=no-value-for-parameter
        thesaurus = event.newParent.__parent__
        if IThesaurus.providedBy(thesaurus):  # pylint: disable=no-value-for-parameter
            index_object(event.object, thesaurus.catalog)


@subscriber(IObjectModifiedEvent, context_selector=IThesaurusTerm)
def handle_modified_term(event):
    """Update index term into inner catalog"""
    parent = get_parent(event.object, IThesaurusTermsContainer)
    if parent is not None:
        thesaurus = parent.__parent__
        if IThesaurus.providedBy(thesaurus):  # pylint: disable=no-value-for-parameter
            reindex_object(event.object, thesaurus.catalog)


@subscriber(IObjectRemovedEvent, context_selector=IThesaurusTerm)
def handle_removed_term(event):
    """Un-index term into inner catalog"""
    parent = event.oldParent
    if IThesaurusTermsContainer.providedBy(parent):  # pylint: disable=no-value-for-parameter
        thesaurus = parent.__parent__
        if IThesaurus.providedBy(thesaurus):  # pylint: disable=no-value-for-parameter
            unindex_object(event.object, thesaurus.catalog)


@adapter_config(required=IThesaurusTerm, provides=INode)
class ThesaurusTermTreeAdapter(ContextAdapter):
    """Thesaurus term tree node adapter"""

    @property
    def label(self):
        """Label getter"""
        return self.context.label

    @property
    def css_class(self):
        """CSS class getter"""
        return self.context.status

    @property
    def order(self):
        """Order getter"""
        return self.context.order or 999

    def get_level(self):
        """Level getter"""
        return self.context.level

    def has_children(self, filter_value=None):
        """Children checker"""
        specifics = self.context.specifics
        if filter_value:
            specifics = list(filter(lambda x: filter_value in (x.extracts or ()), specifics))
        return len(specifics) > 0

    def get_children(self, filter_value=None):  # pylint: disable=unused-argument
        """Children getter"""
        return self.context.specifics


@adapter_config(required=IThesaurusTerm, provides=IViewContextPermissionChecker)
class ThesaurusTermPermissionChecker(ContextAdapter):
    """Thesaurus term permission checker"""

    edit_permission = MANAGE_THESAURUS_CONTENT_PERMISSION


@vocabulary_config(name=THESAURUS_EXTENSIONS_VOCABULARY)
class ThesaurusTermExtensionsVocabulary(SimpleVocabulary):
    """Thesaurus term extensions vocabulary"""

    interface = IThesaurusTermExtension

    def __init__(self, context=None):  # pylint: disable=unused-argument
        registry = get_pyramid_registry()
        translate = check_request().localizer.translate
        try:
            terms = [
                SimpleTerm(name, title=translate(util.label))
                for name, util in registry.getUtilitiesFor(self.interface)
            ]
        except ComponentLookupError:
            terms = []
        super().__init__(terms)



#
# Copyright (c) 2015-2021 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_thesaurus.api.rest module

Thesaurus REST API module.
"""

import sys

from colander import MappingSchema, SchemaNode, SequenceSchema, String, drop
from cornice import Service
from cornice.validators import colander_querystring_validator
from hypatia.text import ParseError
from pyramid.httpexceptions import HTTPOk

from pyams_security.interfaces.base import VIEW_SYSTEM_PERMISSION
from pyams_security.rest import check_cors_origin, set_cors_headers
from pyams_thesaurus.interfaces import REST_EXTRACTS_GETTER_ROUTE, REST_TERMS_SEARCH_ROUTE


__docformat__ = 'restructuredtext'

from pyams_thesaurus import _
from pyams_thesaurus.interfaces.term import STATUS_ARCHIVED
from pyams_thesaurus.interfaces.thesaurus import IThesaurus, IThesaurusExtracts
from pyams_utils.list import unique
from pyams_utils.registry import query_utility


TEST_MODE = sys.argv[-1].endswith('/test')


#
# Thesaurus extracts getter services
#

class ThesaurusExtractsGetterSchema(MappingSchema):
    """Thesaurus extracts getter schema"""
    thesaurus_name = SchemaNode(String(),
                                description=_("Selected thesaurus name"))


class ThesaurusExtractResultSchema(MappingSchema):
    """Thesaurus extracts getter result schema"""


class ThesaurusExtractsResults(SequenceSchema):
    """Thesaurus extracts result schema"""
    result = ThesaurusExtractResultSchema()


class ThesaurusExtractsResultsSchema(MappingSchema):
    """Thesaurus extracts results schema"""
    results = ThesaurusExtractsResults(description=_("Results list"))


extracts_search_response = {
    HTTPOk.code: ThesaurusExtractsResultsSchema(description=_("Search results"))
}
if TEST_MODE:
    extracts_service_params = {}
else:
    extracts_service_params = {
        'response_schemas': extracts_search_response
    }


extracts_service = Service(name=REST_EXTRACTS_GETTER_ROUTE,
                           pyramid_route=REST_EXTRACTS_GETTER_ROUTE,
                           description="Thesaurus extracts management")


@extracts_service.options(validators=(check_cors_origin, set_cors_headers),
                          **extracts_service_params)
def extracts_options(request):  # pylint: disable=unused-argument
    """Extracts service OPTIONS handler"""
    return ''


@extracts_service.get(permission=VIEW_SYSTEM_PERMISSION,
                      schema=ThesaurusExtractsGetterSchema(),
                      validators=(check_cors_origin, colander_querystring_validator,
                                  set_cors_headers),
                      **extracts_service_params)
def get_extracts(request):
    """Get thesaurus extracts list"""
    if TEST_MODE:
        thesaurus_name = request.params.get('thesaurus_name')
    else:
        thesaurus_name = request.validated.get('thesaurus_name')
    if not thesaurus_name:
        return {}
    thesaurus = query_utility(IThesaurus, name=thesaurus_name)
    if thesaurus is None:
        return {}
    extracts = IThesaurusExtracts(thesaurus)
    return {
        'results': [
            {
                'id': extract.name,
                'text': extract.name
            }
            for extract in extracts.values()
        ]
    }


#
# Thesaurus terms search services
#

class ThesaurusTermsSearchQuerySchema(MappingSchema):
    """Thesaurus terms search schema"""
    thesaurus_name = SchemaNode(String(),
                                description=_("Selected thesaurus name"))
    extract_name = SchemaNode(String(),
                              description=_("Selected extract name"),
                              missing=drop)
    term = SchemaNode(String(),
                      description=_("Terms search string"))


class ThesaurusTermResultSchema(MappingSchema):
    """Thesaurus term result schema"""
    id = SchemaNode(String(),
                    description=_("Term ID"))
    text = SchemaNode(String(),
                      description=_("Term label"))


class ThesaurusSearchResults(SequenceSchema):
    """Thesaurus search results interface"""
    result = ThesaurusTermResultSchema()


class ThesaurusSearchResultsSchema(MappingSchema):
    """Thesaurus search results schema"""
    results = ThesaurusSearchResults(description=_("Results list"))


terms_search_response = {
    HTTPOk.code: ThesaurusSearchResultsSchema(description=_("Search results"))
}
if TEST_MODE:
    terms_service_params = {}
else:
    terms_service_params = {
        'response_schemas': terms_search_response
    }


terms_service = Service(name=REST_TERMS_SEARCH_ROUTE,
                        pyramid_route=REST_TERMS_SEARCH_ROUTE,
                        description="Thesaurus terms management")


@terms_service.options(validators=(check_cors_origin, set_cors_headers),
                       **terms_service_params)
def terms_options(request):  # pylint: disable=unused-argument
    """Terms service OPTIONS handler"""
    return ''


@terms_service.get(permission=VIEW_SYSTEM_PERMISSION,
                   schema=ThesaurusTermsSearchQuerySchema(),
                   validators=(check_cors_origin, colander_querystring_validator,
                               set_cors_headers),
                   **terms_service_params)
def get_terms(request):
    """Returns list of terms matching given query"""
    if TEST_MODE:
        thesaurus_name = request.params.get('thesaurus_name')
        extract_name = request.params.get('extract_name')
        query = request.params.get('term')
    else:
        thesaurus_name = request.validated.get('thesaurus_name')
        extract_name = request.validated.get('extract_name')
        query = request.validated.get('term')
    if not (thesaurus_name or query):
        return {}
    thesaurus = query_utility(IThesaurus, name=thesaurus_name)
    if thesaurus is None:
        return {}
    try:
        return {
            'results': [
                {
                    'id': term.label,
                    'text': term.label
                }
                for term in unique(thesaurus.find_terms(query, extract_name,
                                                        exact=True, stemmed=True))
                if term.status != STATUS_ARCHIVED
            ]
        }
    except ParseError:
        return []



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_thesaurus.interfaces main module

This module defines permissions, roles and vocabulary names.
"""


CREATE_THESAURUS_PERMISSION = 'pyams_thesaurus.CreateThesaurus'
'''Permission to create thesaurus'''

ADMIN_THESAURUS_PERMISSION = 'pyams_thesaurus.AdminThesaurus'
'''Permission to administer all thesaurus properties'''

MANAGE_THESAURUS_CONTENT_PERMISSION = 'pyams_thesaurus.ManageThesaurusContent'
'''Permission to manage thesaurus contents'''

MANAGE_THESAURUS_EXTRACT_PERMISSION = 'pyams_thesaurus.ManageThesaurusExtract'
'''Permission to manage thesaurus extract contents'''


THESAURUS_ADMIN_ROLE = 'pyams_thesaurus.Admin'
'''Thesaurus admin is allowed to manage all thesaurus properties'''

THESAURUS_MANAGER_ROLE = 'pyams_thesaurus.Manager'
'''Thesaurus manager is allowed to manage thesaurus terms and extracts'''

THESAURUS_EXTRACT_MANAGER_ROLE = 'pyams_thesaurus.ExtractManager'
'''Thesaurus extract manager is allowed to manage content of a thesaurus extract'''


REST_TERMS_SEARCH_ROUTE = 'thesaurus.rest.terms'
'''Thesaurus terms search route'''

REST_EXTRACTS_GETTER_ROUTE = 'thesaurus.rest.extracts'
'''Thesaurus extracts getter route'''


THESAURUS_NAMES_VOCABULARY = 'pyams_thesaurus.thesaurus'
THESAURUS_EXTRACTS_VOCABULARY = 'pyams_thesaurus.extracts'
